// Generated by CoffeeScript 1.9.3
(function() {
  var QuadTree, Siblings, TILE_SIDE, ZonesManager, debug, defer, lru, lruopts,
    bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  defer = require('node-promise').defer;

  lru = require('lru');

  Siblings = require('./TilerSiblings');

  QuadTree = require('area-qt');

  debug = process.env["DEBUG"];

  TILE_SIDE = 20;

  lruopts = {
    max: 1000,
    maxAgeInMilliseconds: 1000 * 60 * 60 * 24 * 4
  };

  ZonesManager = (function() {
    function ZonesManager(storageEngine, cacheEngine, modelEngine, myAddress, communicationManager, zoneItemQuadTrees, zoneEntityQuadTrees, zoneTiles) {
      this.storageEngine = storageEngine;
      this.cacheEngine = cacheEngine;
      this.modelEngine = modelEngine;
      this.myAddress = myAddress;
      this.communicationManager = communicationManager;
      this.zoneItemQuadTrees = zoneItemQuadTrees;
      this.zoneEntityQuadTrees = zoneEntityQuadTrees;
      this.zoneTiles = zoneTiles;
      this.createNewZone = bind(this.createNewZone, this);
      this.resolveZoneFor = bind(this.resolveZoneFor, this);
      this.lookupZone = bind(this.lookupZone, this);
      this.registerZone = bind(this.registerZone, this);
      this.onSiblingUpdate = bind(this.onSiblingUpdate, this);
      this.shutdown = bind(this.shutdown, this);
      this.onZoneEvicted = bind(this.onZoneEvicted, this);
      this.zoneUnderConstruction = {};
      this.postContructionCallbacks = {};
      this.zones = new lru(lruopts);
      this.zones.on('evict', this.onZoneEvicted);
      this.siblings = new Siblings(this.myAddress, this.communicationManager, this.cacheEngine, this.modelEngine);
      this.communicationManager.registerForUpdates(this.myAddress, this.onSiblingUpdate);
    }

    ZonesManager.prototype.onZoneEvicted = function(zoneObj) {
      return this.siblings.deRegisterAsSiblingForZone(zoneObj);
    };

    ZonesManager.prototype.shutdown = function(zo) {
      return this.siblings.shutdown(zo);
    };

    ZonesManager.prototype.onSiblingUpdate = function(_command, cb) {
      var arg1, arg2, command;
      command = JSON.parse(_command);
      arg1 = command.arg1;
      arg2 = command.arg2;
      switch (command.cmd) {
        case Siblings.CMD_GET_OPLOG:
          return this.siblings.getOplog(command, cb);
        case Siblings.CMD_NEW_OPLOG_EPOCH:
          return this.siblings.newOplogEpoch(command, cb);
      }
    };

    ZonesManager.prototype.registerZone = function(q, zoneObj) {
      var arr, entityQT, itemQT, x, y, ztiles;
      arr = zoneObj.tileid.split('_');
      x = arr[1];
      y = arr[2];
      itemQT = new QuadTree({
        x: x,
        y: y,
        height: TILE_SIDE,
        width: TILE_SIDE
      });
      this.zoneItemQuadTrees[zoneObj.tileid] = itemQT;
      zoneObj.items.forEach((function(_this) {
        return function(item) {
          return _this._setSomething(level, item, itemQT, 'items', q, true).then(function(zo) {});
        };
      })(this));
      entityQT = new QuadTree({
        x: x,
        y: y,
        height: TILE_SIDE,
        width: TILE_SIDE
      });
      this.zoneEntityQuadTrees[zoneObj.tileid] = entityQT;
      zoneObj.entities.forEach((function(_this) {
        return function(entity) {
          return _this._setSomething(level, entity, entityQT, 'entities', q, true).then(function(zo) {});
        };
      })(this));
      ztiles = this.zoneTiles[zoneObj.tileid] || {};
      zoneObj.tiles.forEach((function(_this) {
        return function(tile) {
          return ztiles[tile.x + '_' + tile.y] = tile;
        };
      })(this));
      this.zoneTiles[zoneObj.tileid] = ztiles;
      if (debug) {
        console.log('Tiler.registerZone adds item and entity QTs for tileid ' + zoneObj.tileid);
      }
      this.zones.set(zoneObj.tileid, zoneObj);
      return this.siblings.registerAsSiblingForZone(zoneObj).then((function(_this) {
        return function() {
          var cbs;
          if (_this.zoneUnderConstruction[zoneObj.tileid] === true) {
            delete _this.zoneUnderConstruction[zoneObj.tileid];
            cbs = _this.postContructionCallbacks[zoneObj.tileid] || [];
            cbs.forEach(function(cb) {
              if (debug) {
                console.log('<------ resolving paused lookup of zone');
              }
              return cb();
            });
            return q.resolve(zoneObj);
          } else {
            return q.resolve(zoneObj);
          }
        };
      })(this));
    };

    ZonesManager.prototype.lookupZone = function(tileid, q) {
      var lruZone;
      lruZone = this.zones.get(tileid);
      if (lruZone) {
        return q.resolve(lruZone);
      } else {
        this.zoneUnderConstruction[tileid] = true;
        return this.cacheEngine.get('Zone', tileid).then((function(_this) {
          return function(exists) {
            if (exists) {
              return _this.storageEngine.find('Zone', 'tileid', tileid).then(function(zoneObj) {
                if (debug) {
                  console.log('Tiler.lookupZone got back zone obj');
                }
                if (zoneObj) {
                  if (debug) {
                    console.log('Tiler.lookupZone resolving ' + tileid + ' from db');
                  }
                  return _this.registerZone(q, zoneObj);
                } else {
                  if (debug) {
                    console.log('** Tiler Could not find supposedly existing zone ' + tileid + ' !!!!!');
                  }
                  return q.reject(BAD_TILE);
                }
              });
            } else {
              if (debug) {
                console.log('Tiler.lookupZone zone ' + tileid + ' ****************** not found, so creating new..');
              }
              return _this.createNewZone(tileid).then(function(zoneObj) {
                return _this.registerZone(q, zoneObj);
              });
            }
          };
        })(this));
      }
    };

    ZonesManager.prototype.resolveZoneFor = function(level, x, y) {
      var cbs, q, tid, underConstruction;
      q = defer();
      tid = this.getZoneIdFor(level, x, y);
      underConstruction = this.zoneUnderConstruction[tid];
      if (underConstruction) {
        if (debug) {
          console.log('------> waiting for zone construction for ' + tid);
        }
        cbs = this.postContructionCallbacks[tid] || [];
        cbs.push((function(_this) {
          return function() {
            return _this.lookupZone(tid, q);
          };
        })(this));
        this.postContructionCallbacks[tid] = cbs;
      } else {
        this.lookupZone(tid, q);
      }
      return q;
    };

    ZonesManager.prototype.createNewZone = function(tileid) {
      var newzone, q;
      q = defer();
      newzone = {
        name: 'Zone_' + tileid,
        type: 'Zone',
        id: tileid,
        tileid: tileid,
        items: [],
        entities: [],
        tiles: []
      };
      this.modelEngine.createZone(newzone).then((function(_this) {
        return function(zoneObj) {
          return zoneObj.serialize().then(function() {
            _this.zones.set(tileid, zoneObj);
            return q.resolve(zoneObj);
          });
        };
      })(this));
      return q;
    };

    ZonesManager.prototype.getZoneIdFor = function(level, x, y) {
      var xr, yr, zx, zy;
      xr = x % TILE_SIDE;
      yr = y % TILE_SIDE;
      zx = x - xr;
      zy = y - yr;
      return level + '_' + zx + '_' + zy;
    };

    return ZonesManager;

  })();

  module.exports = ZonesManager;

}).call(this);

//# sourceMappingURL=ZonesManager.js.map
